#include "pktreceiver.h"
int aprobereceiverHdr::offset_;
static class aprobereceiverHdrClass: public PacketHeaderClass
{
        public:
        aprobereceiverHdrClass(): PacketHeaderClass("PacketHeader/aprobe_rcvr",sizeof(aprobereceiverHdrClass))
        {
                bind_offset(&aprobereceiverHdr::offset_);
        }

}class_aprobereceiverHdr;

static class aprobereceiverClass: public TclClass
{
	public:
		aprobereceiverClass():TclClass("Agent/aprobe_rcvr"){}
		TclObject* create(int,const char*const*)
		{
			return (new aprobereceiver());
		}
}class_aprobereceiver;

aprobereceiver::aprobereceiver():Agent(PT_APROBERCVR)
{
	bind("interval_",&interval_);
	bind("NumPacks_",&NumPacks_);
	bind("packetSize_",&size_);
	// This has no role so far
	bind("samples_for_ab_",&samples_for_ab_);  
	bind("decrease_factor_",&decrease_factor_);
	bind("Excursion_Threshold_",&Excursion_Threshold_);
	bind("BinWidth_",&BinWidth_);
	bind("GapDiffIgnore_",&GapDiffIgnore_);
	bind("Resolution_",&Resolution_);
	fleetcounter=0;
	CurrentAb=0;
	state=1;
}

int aprobereceiver::command(int argc,const char*const* argv)
{
	if(argc==2)
	{
		if(strcmp(argv[1],"start")==0)
		{
			start();
			return(TCL_OK);
		}
	}
	return(Agent::command(argc,argv));
}

int aprobereceiver::start()
{
	packetSize_=size_;
	allocatemem();
}

void aprobereceiver::recv(Packet *pkt,Handler*)
{
	hdr_ip* hdrip = hdr_ip::access(pkt);
	aprobesenderHdr *h=aprobesenderHdr::access(pkt);
	int index=(int)h->packetnumber_;
 	double spread=h->SpreadFactor_;
	double interval_=h->interval_;
	int spacketsize=h->packetSize_;
	senderPacks_=h->NumPacks_;
	LinkCapacity=h->Max_rate_;
	int ptype=h->type;
	// This is the data collection part
	if(h->packetnumber_==0)
	{
		rcv_time[index]=Scheduler::instance().clock();	
		snd_time[index]=h->sendtime_;
		qing_delay[index]=rcv_time[index]-(h->sendtime_);
	}
	else
	{
		snd_time[index]=h->sendtime_;
		rcv_time[index]=Scheduler::instance().clock();	
		qing_delay[index]=rcv_time[index]-(h->sendtime_);	
		iat[index]=qing_delay[index]-qing_delay[index-1];
		rates[index]= (double)((8*spacketsize)/(interval_))/1000000;
	}
	// End data collection part
	
	//Inference Logic Part here 
	//One entire probe stream recd
	double feedback;
	if(h->packetnumber_==(senderPacks_-1))
	{
		//Three stage procedure
		//First Stage Compute the Bottleneck Bandwidth
		//Single Probe stream to do that
		//Next Sending Rate is previous receiving rate
		if(state==1)
		{
			feedback=computeBottleneck(spacketsize);
			UpperRate=feedback;
			LowerRate=0;	
			BottleneckBW=feedback;
			CurrentRate=feedback;
			fprintf(stdout,"Bottleneck BW is %lf \n",feedback);
			sendresponse(feedback,1);	
			state=2;
		}
		//Second Stage
		// Reduce the lower rate by two till avr_src_gap and avr_dst gaps are the same
		else if(state==2)
		{
			//Compute Average Src Gap
			avr_src_gap=avrSrcGap();
			//Compute Average Dst Gap
			avr_dst_gap=avrDstGap();
			
			fprintf(stdout,"State2 Upper %lf Lower %lf  ",UpperRate,LowerRate);
			fprintf(stdout,"src %1.9lf and dst %1.9lf\n",avr_src_gap,avr_dst_gap);
			
			//BW estimation procedure done
			double timeofMeasure=rcv_time[index]=Scheduler::instance().clock();	
			// Here the proper Inference Part needs to come
			// Find out the competing CT rate
			/* Here for Competing Traffic */
			double b_gap=(spacketsize*8)/BottleneckBW;
			double src_gap=(spacketsize*8)/CurrentRate;
			double m_gap;
			if(b_gap>src_gap)
			{
				fprintf(stdout," Bgap chosen %1.9lf \n",b_gap);
				m_gap=b_gap; 
			}	
			else
			{
				fprintf(stdout," Srcgap chosen %1.9lf\n",src_gap);
				m_gap=src_gap;
			}
			double gap_diff_sum=0;
			double gap_sum=0,gap_inc=0;
			for(int i=1;i<senderPacks_;i++)
			{
				double cur_gap=rcv_time[i]-rcv_time[i-1];
		   		gap_sum=cur_gap+gap_sum;
				gap_diff_sum+=(cur_gap-m_gap);	
				fprintf(stdout,"Sat %1.9lf and curgap %1.9lf m_gap %1.9lf",gap_inc,cur_gap,m_gap);
				fprintf(stdout," g_sum %1.9lf g_diff_s %1.9lf\n",gap_sum,gap_diff_sum);
			}
			double MGapBW=(spacketsize*8)/(m_gap);
			/*Trial Part here Only */
			if(gap_diff_sum>0)
			{
				double Cbw=(gap_diff_sum*MGapBW)/(m_gap*(senderPacks_-1));
				double nAbw=(MGapBW-Cbw);
				//fprintf(stdout," New AB is %lf \n",nAbw);
				fprintf(stdout,"%lf Bo %lf Mgap %lf ",timeofMeasure,BottleneckBW,MGapBW);
				fprintf(stdout," Co %lf Ab %lf \n",Cbw,nAbw);
			}
			else
			{
				double Cbw=0;
				double nAbw=(MGapBW-Cbw);
				fprintf(stdout,"%lf Bo %lf Mgap %lf ",timeofMeasure,BottleneckBW,MGapBW);
				fprintf(stdout," Co %lf Ab %lf \n",Cbw,nAbw);
			}	
			/*****End Trail Part******/
			state=1;
			fleetcounter=0;
			CurrentAb=0;
			sendresponse(-1,1);
			reset();
		}	
		//freemem();
	}
	Packet::free(pkt);
}


//Study in order packets.. compute the rcv times ..bin them into 25us bins..
//Find the bin with the largest size and include the two adjacent bins.. to get the 
//complete range.. or values. find the average of all the values.. that lie in this range
//This values is the bottleneck link capacity. 
double aprobereceiver::computeBottleneck(int psize)
{
	int numberofbins=0;
	fprintf(stdout," Step 1\n");	
	//Doing the binning procedure
	for(int i=1;i<senderPacks_;i++)
	{
		gaps[i]=rcv_time[i]-rcv_time[i-1];
	   	fprintf(stdout," Gaps are %1.9lf \n",gaps[i]);
		//gaps[i]+=.000008;
		int index=(int)(gaps[i]/BinWidth_);  // this is the index of the bin
	   	fprintf(stdout," Index is %d Bin %1.9lf \n",index,BinWidth_);
		int j=0;
		while((j<numberofbins)&&(value[j]!=index)) j++;
		//New Entry
		if(j==numberofbins)
		{
			value[j]=index;
			count[j]=1;
			numberofbins++;
		}
		//Existing Entry
		else 
		{
			count[j]++;
		}
	}
	
	//Look for the bin with the largest number of elements.
	int max_index=0;
	int max_count=0;
	for(int i=0;i<numberofbins;i++)
	{
		if(count[i]>max_count)
		{
			max_index=i;
			max_count=(int)count[i];
		}
	}
		
	double upper=(value[max_index]+1)*BinWidth_;
	double lower=(value[max_index])*BinWidth_;	
	
	//Get Range of values
	for(int i=0;i<numberofbins;i++)
	{	
		if(value[i]==value[max_index]-1)
			lower=(value[max_index]-1)*BinWidth_;
		if(value[i]==value[max_index]+1)
			upper=(value[max_index]+2)*BinWidth_;
	}
	
	//Now Compute the Bottleneck BW
	double gap_sum=0;
	double gap_count=0;
	for(int i=0;i<senderPacks_;i++)
	{
		if((gaps[i]>=lower)&&(gaps[i]<=upper))
		{
			gap_sum=gap_sum+gaps[i];
			gap_count++;
		}
	}
	//double bbw=(8*packetSize_)/(gap_sum/gap_count);
	double bbw=(8*psize)/(gap_sum/gap_count);
	fprintf(stdout," Step 5\n");	
	return bbw;
}

double aprobereceiver::Converge()
{

}

double aprobereceiver::Comp(double d,double s)
{
	
	//if(d>s+GapDiffIgnore_)
	if(d>(s*(1+GapDiffIgnore_)))
	{
		return(1);
	}
	else
	{
		return (0);
	}
}

double aprobereceiver::avrSrcGap()
{
	double sum=0;
	int ctr=0;
	for(int i=1;i<senderPacks_;i++)
	{
		double gap=snd_time[i]-snd_time[i-1];
		sum=sum+gap;
		ctr++;
	}
	double av=sum/ctr;
	return av;	
}
double aprobereceiver::avrDstGap()
{
	double sum=0;
	int ctr=0;
	for(int i=1;i<senderPacks_;i++)
	{
		double gap=rcv_time[i]-rcv_time[i-1];
		sum=sum+gap;
		ctr++;
	}
	double av=sum/ctr;
	return av;	
}

void aprobereceiver::allocatemem()
{
	int tempsenderPacks=10000;
	int Esqrt=(int)sqrt(tempsenderPacks);
	qing_delay=(double *) calloc((int)(tempsenderPacks),sizeof(double));
	rates=(double *) calloc((int)(tempsenderPacks),sizeof(double));
	iat=(double *) calloc((int)(tempsenderPacks),sizeof(double));
	av_bw_per_pkt=(double *) calloc((int)(tempsenderPacks),sizeof(double));
	snd_time=(double *) calloc((int)(tempsenderPacks),sizeof(double));
	rcv_time=(double *) calloc((int)(tempsenderPacks),sizeof(double));
	gaps=(double *) calloc((int)(tempsenderPacks),sizeof(double));
	value=(double*) calloc((int)(tempsenderPacks),sizeof(double));
	count=(double*) calloc((int)(tempsenderPacks),sizeof(double));
}	

void aprobereceiver::freemem()
{
	free(qing_delay);
	free(rates);
	free(iat);
	free(av_bw_per_pkt);
	free(snd_time);
	free(rcv_time);
}

double aprobereceiver::computeadr(int spacketsize)
{
	int last=(int)(senderPacks_-1);
	double delta=rcv_time[last]-rcv_time[0];
	double adr=(spacketsize*8*senderPacks_)/delta;
	return adr;	 	
}

void aprobereceiver::sendresponse(double newrate,int type)
{
   Packet *p=allocpkt();
   aprobereceiverHdr *hdr=aprobereceiverHdr::access(p);
   hdr->sendtime_=Scheduler::instance().clock();
	hdr->ack_=0;
	hdr->newPacketSize_=1500;   // This will change
	hdr->responseType_=type;    
	hdr->feedback_=newrate;
	hdr->loss_=0;              // This will change
	hdr->terminate_=0; 	   // This will change
   send(p,0);
   return;
}

double aprobereceiver::computeavabw()
{
	double deltain=snd_time[(int)(senderPacks_-1)]-snd_time[0];
	double deltaout=rcv_time[(int)(senderPacks_-1)]-rcv_time[0];
	double ratio=(deltaout-deltain)/(deltain);
	double instabw=LinkCapacity*(1-ratio);	
	CurrentAb=CurrentAb+instabw;
	fprintf(stdout,"deltain %lf and deltaout %lf \n",deltain,deltaout);
}

void aprobereceiver::reset()
{
	fleetcounter=0;
	CurrentAb=0;
	state=1;
	BottleneckBW=0;
}
