#include "pktreceiver.h"

int pktreceiverHdr::offset_;
static class pktreceiverHdrClass: public PacketHeaderClass
{
	public:
	pktreceiverHdrClass(): PacketHeaderClass("PacketHeader/pktreceiver",sizeof(pktreceiverHdrClass))
	{
		bind_offset(&pktreceiverHdr::offset_);
	}

}class_pktreceiverHdr;

static class pktreceiverClass: public TclClass
{
	public:
		pktreceiverClass():TclClass("Agent/pktreceiver"){}
		TclObject* create(int,const char*const*)
		{
			return (new pktreceiver());
		}
}class_pktreceiver;

pktreceiver::pktreceiver():Agent(PT_PKTRECEIVER)
{
	bind("interval_",&interval_);
	bind("NumPacks_",&NumPacks_);
	bind("packetSize_",&size_);
	bind("decrease_factor_",&decrease_factor_);
	bind("samples_for_ab_",&samples_for_ab_);
	bind("Excursion_Threshold_",&Excursion_Threshold_);
	full=0;
	head=0;
	tail=0;
	samples=0;
}

int pktreceiver::command(int argc,const char*const* argv)
{
	if(argc==2)
	{
		if(strcmp(argv[1],"start")==0)
		{
			start();
			return(TCL_OK);
		}
	}
	return(Agent::command(argc,argv));
}

int pktreceiver::start()
{
// right now we are not doing anything
// allocate the memory for some data structures
	circular=(double *)calloc((int)(samples_for_ab_),sizeof(double));
}

void pktreceiver::recv(Packet *pkt,Handler*)
{
	hdr_ip* hdrip = hdr_ip::access(pkt);
	pktsenderHdr *h=pktsenderHdr::access(pkt);
	int index=(int)h->packetnumber_;
 	double spread=h->SpreadFactor_;
	double interval_=h->interval_;
	int spacketsize=h->packetSize_;
	senderPacks_=h->NumPacks_;
	//fprintf(stdout,"RAll1 %lf %d %lf \n",h->sendtime_,h->packetSize_,h->packetnumber_);
   	fprintf(stdout,"Packet Number %d \n",index);
	if(h->packetnumber_==0)
	{
		allocatemem();
		qing_delay[index]=(Scheduler::instance().clock())-(h->sendtime_);	
		fprintf(stdout," First Quing %lf \n",qing_delay[index],iat[index]);
	}
	else
	{
		qing_delay[index]=(Scheduler::instance().clock())-(h->sendtime_);	
		//iat[index]=qing_delay[index]-qing_delay[index-1];
		double decrease=1;
      		for(int i=0;i<(index-1);i++)
      		{ 
      			decrease=decrease*spread;
      		}	  
		iat[index]=(interval_*decrease);
		rates[index]= (double)((8*spacketsize)/(interval_*decrease))/1000000;
		
		fprintf(stdout,"Spread %1.9lf Quing %1.18lf \n",spread,qing_delay[index]);
		fprintf(stdout,"Rates %1.9lf IAT %1.18lf \n",rates[index],iat[index]);
	}
	if(h->packetnumber_==(senderPacks_))
	{
		fprintf(stdout," Compute Called \n");
		double inst_abw=computeavabw();
		double time=Scheduler::instance().clock();
		fprintf(stdout," %lf: InstAvailable Bandwidth is %lf \n",time,inst_abw);
		freemem();
		fprintf(stdout," Alok %d %d \n",tail,samples_for_ab_);
		circular[tail]=inst_abw;
		samples++;
		tail++;
		if(tail>samples_for_ab_) tail=0;
		if(samples>(samples_for_ab_-1))
		{
			double sum,MovAB;	
			for(int i=0;i<samples_for_ab_;i++)
			{
				//fprintf(stdout," Puts %lf ",circular[i]);
				sum=sum+circular[i];
			}
			double MovAb=sum/samples_for_ab_;
			double time=Scheduler::instance().clock();
			fprintf(stdout,"%lf Moving Available BW is %lf \n",time,MovAb);
		}
	}
	Packet::free(pkt);
}

void pktreceiver::allocatemem()
{
	fprintf(stdout,"In allocate Mem %d \n",NumPacks_);
	qing_delay=(double *) calloc((int)(NumPacks_),sizeof(double));
	rates=(double *) calloc((int)(NumPacks_),sizeof(double));
	iat=(double *) calloc((int)(NumPacks_),sizeof(double));
	av_bw_per_pkt=(double *) calloc((int)(NumPacks_),sizeof(double));
}	

void pktreceiver::freemem()
{
	fprintf(stdout,"In free Mem %d \n",NumPacks_);
	free(qing_delay);
	free(rates);
	free(iat);
	free(av_bw_per_pkt);
}	

double pktreceiver::computeavabw()
{
	int loc=0,exstart=0,exend=0;
	double sum_iat, sum_abw;
	double maxq=0;
	sum_abw=0;
	sum_iat=0;
	//Look for first excursion
	while((qing_delay[loc]>=(qing_delay[loc+1]))&&(loc<NumPacks_))
		loc++;
	exstart=loc;
	loc++;
	fprintf(stdout," First Location Found %d \n",exstart); 
	while(loc<senderPacks_)
	{
		if(maxq<(qing_delay[loc+1]-qing_delay[exstart]))
			maxq=qing_delay[loc+1]-qing_delay[loc];
		// Chk to see if the qing delay has decreased	
		if((qing_delay[loc]-qing_delay[exstart])<(maxq/decrease_factor_))
		{
			//fprintf(stdout," Inside loop \n");
			// If we enter this condition.. the excursion has ended
			// If greater than this threshold then worthwhile 
			// Else might be noise
			if((loc-exstart) >= Excursion_Threshold_)
			{
				//Mark all increasing periods as having AB equal to their
				//Rates
				for(int count=exstart;count<loc-1;count++)
				{
					if(qing_delay[count]<qing_delay[count+1])
						av_bw_per_pkt[count]=rates[count];
				}
			}
			while((qing_delay[loc]>=qing_delay[loc+1])&&(loc<NumPacks_))
				loc++;	
			exstart=loc;
		}
		loc++;
	}
	//For the last excursion let all be the same as the sending rate for the first
	if(exstart==senderPacks_)
		exstart--;
	for(int counter=exstart;counter<senderPacks_;counter++)
	{
		//fprintf(stdout," Setting Rates\n");
		av_bw_per_pkt[counter]=rates[exstart];
	}
	//Compute the AB from the rates equation
	for(int count=1;count<senderPacks_;count++)
	{
		if(av_bw_per_pkt[count]<1.0)
			av_bw_per_pkt[count]=rates[exstart];
		sum_iat=sum_iat+iat[count];
		sum_abw=sum_abw+av_bw_per_pkt[count]*iat[count];
		fprintf(stdout," Abw per packet is %lf %lf \n",av_bw_per_pkt[count],iat[count]);
	}
	fprintf(stdout,"Sum IAT %lf and Sum ABW %lf \n",sum_iat, sum_abw);
	sum_abw=sum_abw/sum_iat;
	return sum_abw;
}		
